package tester;

import java.lang.reflect.*;


/**
 * <P>A test harness that compares arbitrary objects for 
 * extensional equality. </P>
 * <P>It catches exceptions in tests - the remaining tests are omitted.</P>
 * <P>It uses the visitor pattern to accept test cases.</P>
 * 
 * @author Viera K. Proulx
 * @since 3 June 2007
 */
public class Tester{

  /** A <code>String</code> that records the results for all failed tests */
  String testResults = "Test results: \n";

  /** A <code>String</code> that records all test results */
  String fullTestResults = "Full test results: \n";

  int n;			    /** the total number of tests */
  int errors;		    /** the total number of errors */
  String testname;      /** the name of the current test
                            to report on Runtime exception */

  /** start with no tests and no failures */
  public Tester() {
    this.n = 0; 
    this.errors = 0;
    this.testname = "";
  }

  /**
   * Run the tests, accept the class to be tested as a visitor
   * 
   * @param f the class to be tested -- it defines the method 
   * <code>void tests(Tester t)</code>).
   */
  public void runTests(Testable f) {
    this.n = 0;
    boolean failed = false;

    try {
      f.tests(this);
    }
    catch (Throwable e) {  // catch all exceptions
      this.errors = this.errors + 1;
      System.out.println("Threw exception during test " + this.n);
      e.printStackTrace();
      failed = true;
    }
    finally {
      done(failed);
    }
  }

  /**
   * On completion of all tests print the count of number of tests 
   * and the number of failures.
   */
  public void done(boolean failed){
    if (failed){
      this.reportErrors(testname, "caused RuntimeException", "unkown");
    }
    if (this.errors > 0)
      System.out.print("Failed " + this.errors + " out of ");
    else
      System.out.print("Passed all ");
    System.out.println (this.n + " tests.");
  }

  /**
   * Set the relative tolerance for the comparison of inexact numbers
   * @param epsilon the desired tolerance
   */
  public void setTolerance(double epsilon){
    Inspector.TOLERANCE = epsilon;
  }


  //--------- DELEGATE THE same METHOD TO THE Inspector CLASS --------

  /**
   * <P>Provide a general extensional equality comparison for 
   * arbitrary pair of objects</P>
   * <P>Inspector invokes the user-define <code>same</code> method, if the 
   * corresponding classes implement the <code>ISame</code> interface.</P>
   * 
   * @param obj1 the first object
   * @param obj2 the second object
   * @return true if the two objects represent the same data
   */
  public boolean same(Object obj1, Object obj2){
    return Inspector.same(obj1, obj2);
  }

  //--------- TESTS --------------------------------------------------

  /**
   * General contractor to report test results
   * 
   * @param success did the test succeed?
   * @param testname the name of this test
   * @param actual the computed value of the result
   * @param expected the expected value of the result
   */
  public void report(boolean success, String testname, 
      Object actual, Object expected){
    if (success)
      this.reportSuccess(testname, actual, expected);
    else
      this.reportErrors(testname, actual, expected);
  }


  /**
   * Test that only reports success or failure
   * 
   * @param testname the name of this test
   * @param result the test result
   */
  public void test(String testname, boolean result){
    this.testname = testname;
    if (!result){
      // add the report of failure
      addError(
          testname + ": error -- no blame -- \n");
    }
    // add the report of success
    else{
      this.addSuccess(
          testname + ": success \n");
    }
  }

  /**
   * Test that compares two <code>boolean</code> values using == operator
   * 
   * @param testname the name of this test
   * @param actual the computed boolean value
   * @param expected the expected boolean value
   */
  public void test(String testname, boolean actual, boolean expected){  
    this.testname = testname; 
    this.report(actual == expected, 
        testname, "" + actual, "" + expected);
  }

  /**
   * Test that compares two objects of any kind
   * 
   * @param <T> the type of the objects being compared
   * @param testname the name of this test
   * @param actual the computed value of the type T
   * @param expected the expected value of the type T
   */
  public <T>  void test(String testname, T actual, T expected){ 
    this.testname = testname;
    this.report(Inspector.same(actual, expected), 
        testname, actual, expected);
  }

 
  /**
   * Test that verifies that when the given object invokes the given
   * method with the given arguments, it throws the expected exception
   * with the expected message.
   * 
   * @param <T> The type of the object that invokes the given method
   * @param testname the description of this test
   * @param object the object that invokes the method to be tested
   * @param method the name of the method to test
   * @param args an array of arguments for the method - use Array of length 0
   * if no arguments are needed
   * @param e an instance of the expected exception -- including the 
   * expected message
   */
  public <T>  void test(String testname, T object, String method, 
                          Object[] args, Exception e){ 
    this.testname = testname;
    
    // create an Array of the argument types
    int length = Array.getLength(args);
    Class[] parameters = new Class[length];

    for (int i = 0; i < length; i++){
      parameters[i] = args[i].getClass();
    }
    
    // get the type of the expected exception
    Class exceptClass = e.getClass();
    try {
      
      // invoke the method
      Method meth = object.getClass().getMethod(method, parameters);
      // allow access to protected and private methods
      meth.setAccessible(true);
      
      Object result = meth.invoke(object, args);
      
      // if the invocation succeeds, the test fails because
      // it does not throw the expected exception
      this.addError(testname + " : errors" + 
          "\n invocation did not throw any exception " +
            "\n  method name: " + meth.getName() +
            "\n  object class: " + object.getClass().getName() + 
            "\n  result: " + result.getClass().getName()  +
            ": " + Inspector.makeString(result) + 
            "\n  expected exception was: \n    class: " 
            + e.getClass().getName() + 
            "\n    message: " + e.getMessage() + "\n\n");
    }
    catch (Throwable exception){
      
      // check that the method threw an exception of the desired type
      if (exception.getCause().getClass().getName().equals(
                                             exceptClass.getName())){
        
        // check whether the message is correct
        if (e.getMessage().equals(exception.getCause().getMessage())){
          
          // report correct exception, correct message -- test succeeds
          this.addSuccess(testname + " : success" + 
              "\n correct exception: \n class: " +
              exception.getCause().getClass().getName() +
              "\n correct message: " + 
              exception.getCause().getMessage() + "\n\n");
        }
        else{
          
          // report correct exception, incorrect message -- test fails
          this.addError(testname + " : errors" + 
              "\n correct exception: \n class: " +
              exception.getCause().getClass().getName() +
              "\n incorrect message: " +
              exception.getCause().getMessage() + "\n\n");
        }
      }
      else{
        // report that method invocation threw an exception of the wrong type
        this.addError(testname + " : errors" + 
            "\n incorrect exception was thrown: " +
            "\n    class of the thrown exception: " +
            exception.getCause().getClass().getName() +
            "\n    class of the expected exception: " +
            e.getClass().getName() + "\n\n");
      }
    }
  }


  /**
   * Invoke the method with the given name on a given object
   * with the given arguments - check if it produces the 
   * expected value
   * 
   * @param <T> The type of the object that invokes the given method
   * @param testname the description of this test
   * @param object the object that invokes the method to be tested
   * @param method the name of the method to test
   * @param args an array of arguments for the method - use Array of length 0
   * if no arguments are needed
   * @param expected 
   */
  public <T>  void test(String testname, T object, String method, 
      Object[] args, Object expected){ 
    this.testname = testname;
    // create an Array of the argument types
    int length = Array.getLength(args);
    Class[] parameters = new Class[length];

    for (int i = 0; i < length; i++){
      parameters[i] = args[i].getClass();
    }
    
    try {

      // invoke the method -- it should produce a boolean value
      System.out.println("invoking the method " + method + " in the class " +
          object.getClass().getName());
      Method meth = object.getClass().getDeclaredMethod(method, parameters);
      // allow access to protected and private methods
      meth.setAccessible(true);

      String testmessage = testname + "\n" + 
        Inspector.makeString(object) + "\n invoked method " + method +
        " in the class " + object.getClass().getName() + 
        "\n with arguments " + Inspector.makeString(args);
      
      test(testmessage, meth.invoke(object, args), expected);
    }
    catch (Throwable exception){
      String testmessage = testname + "\n" + 
      Inspector.makeString(object) + "\n invoked method " + method +
      " in the class " + object.getClass().getName() + 
      "\n with arguments " + Inspector.makeString(args);
      
      // check that the method threw an exception of the desired type
      this.reportErrors(testname + testmessage + "\nthrew an excception ", 
          object, Inspector.makeString(args));
      exception.printStackTrace();
    }
        
  
  }
  
  

  //--------- TEST REPORTS ------------------------------------------------

  /**
   * Add a test to the list of failed tests
   * 
   * @param testname the name of the failed test
   * @param actual the computed value of the test
   * @param expected the expected value of the test
   */
  public void reportErrors(String testname, 
      Object actual, 
      Object expected){

    // add test report to the error report and the full test report
    this.addError(
        testname + ": errors \n" +
        "actual: " + Inspector.makeString(actual) + "\n" +
        "expected:   " + Inspector.makeString(expected) + "\n");
  }

  /**
   * Add a test to the list of successful tests
   * 
   * @param testname the name of the successful test
   * @param actual the computed value of the test
   * @param expected the expected value of the test
   */
  public void reportSuccess(String testname, 
      Object actual, 
      Object expected){

    // add test report to the full test report
    this.addSuccess(
        testname + ": success \n" +
        "actual: " + Inspector.makeString(actual) + "\n" +
        "expected:   " + Inspector.makeString(expected) + "\n");
  }

  /**
   * Add the given test successful test result to the full report 
   * @param testResult
   */
  public void addSuccess(String testResult){
    this.fullTestResults = this.fullTestResults + "\n" + testResult;
    // update the count of all tests
    this.n = this.n + 1;
  }

  /**
   * Add the given test failed test result to the full report and 
   * to the error report
   * @param testResult
   */
  public void addError(String testResult){
    this.testResults = this.testResults + "\n" + testResult;
    this.fullTestResults = this.fullTestResults + "\n" + testResult;
    //  update the count of all tests
    this.n = this.n + 1;
    //  update the count of the errors tests
    this.errors = this.errors + 1;    
  }

  /**
   * Report on the number and nature of failed tests
   */
  public void testReport(){
    if (this.errors == 1){
      System.out.println(this.testResults + 
          this.errors + " test failed.\n");
    }
    else if (this.errors > 1){
      System.out.println(this.testResults + 
          this.errors + " tests failed.\n");
    }
    else
      System.out.println(this.testResults + 
      "All tests passed.\n");
  }

  /**
   * Produce test names and values compared for all tests
   */
  public void fullTestReport(){
    if (this.errors == 1){
      System.out.println("\n 1 test failed.\n");
    }
    else if (this.errors > 1){
      System.out.println("\n " + this.errors + " tests failed.\n");
    }
    else
      System.out.println("All tests passed.\n");
    System.out.println(this.fullTestResults);
  }
  
  /**
   * A convenience method for the tester users - to produce a String 
   * representation of the given object.
   * 
   * @param obj the object to represent as a <CODE>String</CODE>
   * @return a String representation of the given object.
   */
  public String makeString(Object obj){
    return Inspector.makeString(obj);
  }

}